﻿using System;
using System.ComponentModel;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;

namespace Inedo.Agents.Hosting
{
    internal sealed class SlimProcess : IDisposable
    {
        private IntPtr handle;
        private bool disposed;
        private ProcessWaitHandle waitHandle;
        private RegisteredWaitHandle registeredWaitHandle;
        private Action<SlimProcess> exited;
        private bool exitNotificationReceived;
        private readonly object notifyLock = new object();

        private SlimProcess(IntPtr handle, int id, Action<SlimProcess> exited)
        {
            this.exited = exited;
            this.handle = handle;
            this.Id = id;
            this.waitHandle = new ProcessWaitHandle(handle);

            lock (notifyLock)
            {
                this.registeredWaitHandle = ThreadPool.RegisterWaitForSingleObject(this.waitHandle, this.HandleProcessExited, null, -1, true);
                if (!this.IsRunning)
                {
                    this.exitNotificationReceived = true;
                    exited(this);
                }
            }
        }

        ~SlimProcess()
        {
            this.Dispose(false);
        }

        public int Id { get; }
        public int ExitCode
        {
            get
            {
                uint exitCode;
                NativeMethods.GetExitCodeProcess(this.handle, out exitCode);
                return (int)exitCode;
            }
        }
        public bool IsRunning => this.ExitCode == 259;

        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        public static SlimProcess Start(string fileName, string messengerId, Action<SlimProcess> exited, ScopedConfig config)
        {
            var commandLine = new StringBuilder(100);
            commandLine.Append('\"');
            commandLine.Append(fileName);
            commandLine.Append("\" --messenger=");
            commandLine.Append(messengerId);
            commandLine.Append(" --pid=");
            commandLine.Append(GetCurrentProcessId());
            commandLine.Append(" --extensionsPath=\"");
            commandLine.Append(config.ExtensionsPath.TrimEnd('\\'));
            commandLine.Append("\" --extensionsTempPath=\"");
            commandLine.Append(config.ExtensionsTempPath.TrimEnd('\\'));
            commandLine.Append("\" --tempPath=\"");
            commandLine.Append(config.TempPath.TrimEnd('\\'));
            commandLine.Append('\"');

            var info = new PROCESS_INFORMATION();
            try
            {
                unsafe
                {
                    var startupInfo = new STARTUPINFO
                    {
                        cb = (uint)sizeof(STARTUPINFO)
                    };

                    if (!NativeMethods.CreateProcess(fileName, commandLine, IntPtr.Zero, IntPtr.Zero, false, 0x8, IntPtr.Zero, null, &startupInfo, &info))
                        throw new Win32Exception();

                    return new SlimProcess(info.hProcess, (int)info.dwProcessId, exited);
                }
            }
            catch
            {
                if (info.hProcess != IntPtr.Zero)
                    NativeMethods.CloseHandle(info.hProcess);

                throw;
            }
            finally
            {
                if (info.hThread != IntPtr.Zero)
                    NativeMethods.CloseHandle(info.hThread);
            }
        }
        public static SlimProcess Start(string fileName, string arguments)
        {
            var commandLine = new StringBuilder(100);
            commandLine.Append('\"');
            commandLine.Append(fileName);
            commandLine.Append("\" ");
            commandLine.Append(arguments);

            var info = new PROCESS_INFORMATION();
            try
            {
                unsafe
                {
                    var startupInfo = new STARTUPINFO
                    {
                        cb = (uint)sizeof(STARTUPINFO)
                    };

                    if (!NativeMethods.CreateProcess(fileName, commandLine, IntPtr.Zero, IntPtr.Zero, false, 0x8, IntPtr.Zero, null, &startupInfo, &info))
                        throw new Win32Exception();

                    return new SlimProcess(info.hProcess, (int)info.dwProcessId, p => { });
                }
            }
            catch
            {
                if (info.hProcess != IntPtr.Zero)
                    NativeMethods.CloseHandle(info.hProcess);

                throw;
            }
            finally
            {
                if (info.hThread != IntPtr.Zero)
                    NativeMethods.CloseHandle(info.hThread);
            }
        }

        private void HandleProcessExited(object state, bool timedOut)
        {
            lock (this.notifyLock)
            {
                if (!this.exitNotificationReceived)
                {
                    this.exitNotificationReceived = true;
                    this.exited(this);
                }
            }
        }

        private void Dispose(bool disposing)
        {
            if (!this.disposed)
            {
                if (disposing)
                {
                    this.registeredWaitHandle.Unregister(null);
                    this.registeredWaitHandle = null;

                    this.waitHandle.Dispose();
                    this.waitHandle = null;
                }

                if (this.handle != IntPtr.Zero)
                {
                    NativeMethods.CloseHandle(this.handle);
                    this.handle = IntPtr.Zero;
                }

                this.disposed = true;
            }
        }

        private static int GetCurrentProcessId()
        {
            using (var process = System.Diagnostics.Process.GetCurrentProcess())
            {
                return process.Id;
            }
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct PROCESS_INFORMATION
        {
            public IntPtr hProcess;
            public IntPtr hThread;
            public uint dwProcessId;
            public uint dwThreadId;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct STARTUPINFO
        {
            public uint cb;
            public IntPtr lpReserved;
            public IntPtr lpDesktop;
            public IntPtr lpTitle;
            public uint dwX;
            public uint dwY;
            public uint dwXSize;
            public uint dwYSize;
            public uint dwXCountChars;
            public uint dwYCountChars;
            public uint dwFillAttribute;
            public uint dwFlags;
            public ushort wShowWindow;
            public ushort cbReserved2;
            public IntPtr lpReserved2;
            public IntPtr hStdInput;
            public IntPtr hStdOutput;
            public IntPtr hStdError;
        }

        private static class NativeMethods
        {
            [return: MarshalAs(UnmanagedType.Bool)]
            [DllImport("kernel32.dll", CallingConvention = CallingConvention.Winapi, CharSet = CharSet.Unicode, SetLastError = true)]
            public static extern unsafe bool CreateProcess(
                [MarshalAs(UnmanagedType.LPWStr)] string lpApplicationName,
                [MarshalAs(UnmanagedType.LPWStr)] StringBuilder lpCommandLine,
                IntPtr lpProcessAttributes,
                IntPtr lpThreadAttributes,
                [MarshalAs(UnmanagedType.Bool)] bool bInheritHandles,
                uint dwCreationFlags,
                IntPtr lpEnvironment,
                [MarshalAs(UnmanagedType.LPWStr)] string lpCurrentDirectory,
                STARTUPINFO* lpStartupInfo,
                PROCESS_INFORMATION* lpProcessInformation
            );

            [return: MarshalAs(UnmanagedType.Bool)]
            [DllImport("kernel32.dll", CallingConvention = CallingConvention.Winapi, SetLastError = true)]
            public static extern bool CloseHandle(IntPtr hObject);

            [return: MarshalAs(UnmanagedType.Bool)]
            [DllImport("kernel32.dll", CallingConvention = CallingConvention.Winapi, SetLastError = true)]
            public static extern bool GetExitCodeProcess(IntPtr hProcess, out uint lpExitCode);
        }

        private sealed class ProcessWaitHandle : WaitHandle
        {
            public ProcessWaitHandle(IntPtr handle)
            {
                this.SafeWaitHandle = new Microsoft.Win32.SafeHandles.SafeWaitHandle(handle, false);
            }
        }
    }
}
