﻿using System;
using System.Linq;
using System.Reflection;
using System.ServiceProcess;
using System.Threading.Tasks;

namespace Inedo.WindowsServices
{
    /// <summary>
    /// Base class for Inedo services.
    /// </summary>
    public abstract class InedoService
    {
        private readonly object stateLock = new object();
        private TaskCompletionSource<object> taskCompletionSource;
        private bool isPaused;

        /// <summary>
        /// Initializes a new instance of the <see cref="InedoService"/> class.
        /// </summary>
        protected InedoService()
        {
        }

        /// <summary>
        /// Gets the name of the service.
        /// </summary>
        public abstract string Name { get; }
        /// <summary>
        /// Gets the display name of the service.
        /// </summary>
        public virtual string FriendlyName
        {
            get
            {
                var title = AhReflection.GetCustomAttribute<AssemblyTitleAttribute>(this.GetType().Assembly);
                return title?.Title ?? this.Name;
            }
        }
        /// <summary>
        /// Gets the description of the service.
        /// </summary>
        public virtual string Description
        {
            get
            {
                var description = AhReflection.GetCustomAttribute<AssemblyDescriptionAttribute>(this.GetType().Assembly);
                return description?.Description ?? this.FriendlyName;
            }
        }
        /// <summary>
        /// Gets a value indicating whether the service supports pause and continue commands.
        /// </summary>
        public virtual bool CanPauseAndContinue => false;
        /// <summary>
        /// Gets a task which completes when the service is stopped.
        /// </summary>
        public Task Task { get; private set; }
        /// <summary>
        /// Gets a value indicating whether the service is currently running.
        /// </summary>
        public bool IsRunning { get; private set; }

        /// <summary>
        /// Starts the service.
        /// </summary>
        /// <exception cref="InvalidOperationException">The service is already running.</exception>
        public void Start()
        {
            lock (this.stateLock)
            {
                if (this.IsRunning)
                    throw new InvalidOperationException("The service is already running.");

                this.OnStart();
                this.IsRunning = true;
                this.taskCompletionSource = new TaskCompletionSource<object>();
                this.Task = taskCompletionSource.Task;
            }
        }
        /// <summary>
        /// Stops the service.
        /// </summary>
        /// <exception cref="InvalidOperationException">The service is already stopped.</exception>
        public void Stop()
        {
            lock (this.stateLock)
            {
                if (!this.IsRunning)
                    throw new InvalidOperationException("The service is already stopped.");

                try
                {
                    this.OnStop();
                }
                catch (Exception ex)
                {
                    this.taskCompletionSource.TrySetException(ex);
                    this.taskCompletionSource = null;
                    return;
                }
                finally
                {
                    this.IsRunning = false;
                }

                this.taskCompletionSource.TrySetResult(null);
                this.taskCompletionSource = null;
            }
        }
        /// <summary>
        /// Pauses the service.
        /// </summary>
        /// <exception cref="InvalidOperationException">The service is already stopped or paused.</exception>
        /// <exception cref="NotSupportedException">The service does not accept pause commands.</exception>
        public void Pause()
        {
            lock (this.stateLock)
            {
                if (!this.IsRunning)
                    throw new InvalidOperationException("The service is already stopped.");
                if (this.isPaused)
                    throw new InvalidOperationException("The service is already paused.");
                if (!this.CanPauseAndContinue)
                    throw new NotSupportedException("The service does not accept pause commands.");

                this.OnPause();
                this.isPaused = true;
            }
        }
        /// <summary>
        /// Continues the service.
        /// </summary>
        /// <exception cref="InvalidOperationException">The service is already stopped or running.</exception>
        /// <exception cref="NotSupportedException">The service does not accept continue commands.</exception>
        public void Continue()
        {
            lock (this.stateLock)
            {
                if (!this.IsRunning)
                    throw new InvalidOperationException("The service is already stopped.");
                if (!this.isPaused)
                    throw new InvalidOperationException("The service is already running.");
                if (!this.CanPauseAndContinue)
                    throw new NotSupportedException("The service does not accept pause commands.");

                this.OnContinue();
                this.isPaused = false;
            }
        }

        /// <summary>
        /// Returns a <see cref="ServiceBase"/> which wraps this instance.
        /// </summary>
        /// <returns>A <see cref="ServiceBase"/> which wraps this instance.</returns>
        public ServiceBase GetWindowsService() => new InedoWindowsService(this);
        /// <summary>
        /// Returns a <see cref="ServiceBase"/> which wraps the specified services.
        /// </summary>
        /// <param name="services">The services to wrap.</param>
        /// <returns>A <see cref="ServiceBase"/> which wraps the specified services.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="services"/> is null.</exception>
        /// <exception cref="ArgumentException"><paramref name="services"/> is empty.</exception>
        public static ServiceBase GetWindowsService(params InedoService[] services)
        {
            if (services == null)
                throw new ArgumentNullException(nameof(services));
            if (services.Length == 0)
                throw new ArgumentException("At least one service must be specified.");

            if (services.Length == 1)
                return new InedoWindowsService(services[0]);
            else
                return new CombinedInedoWindowsService(services.ToArray());
        }

        /// <summary>
        /// Invoked when the service receives a start command.
        /// </summary>
        protected abstract void OnStart();
        /// <summary>
        /// Invoked when the service receives a stop command.
        /// </summary>
        protected abstract void OnStop();
        /// <summary>
        /// Invoked when the service receives a pause command.
        /// </summary>
        protected virtual void OnPause()
        {
        }
        /// <summary>
        /// Invoked when the service receives a continue command.
        /// </summary>
        protected virtual void OnContinue()
        {
        }

        private sealed class InedoWindowsService : ServiceBase
        {
            private InedoService service;

            public InedoWindowsService(InedoService service)
            {
                this.service = service;
                this.CanPauseAndContinue = service.CanPauseAndContinue;
                this.ServiceName = service.Name;
            }

            protected override void OnStart(string[] args)
            {
                this.service.Start();
            }
            protected override void OnStop()
            {
                this.service.Stop();
            }
            protected override void OnPause()
            {
                this.service.OnPause();
            }
            protected override void OnContinue()
            {
                this.service.OnContinue();
            }
        }

        private sealed class CombinedInedoWindowsService : ServiceBase
        {
            private InedoService[] services;

            public CombinedInedoWindowsService(InedoService[] services)
            {
                this.services = services;
                this.CanPauseAndContinue = services.All(s => s.CanPauseAndContinue);
            }

            protected override void OnStart(string[] args)
            {
                foreach (var service in this.services)
                    service.Start();
            }
            protected override void OnStop()
            {
                foreach (var service in this.services)
                    service.Stop();
            }
            protected override void OnPause()
            {
                foreach (var service in this.services)
                    service.Pause();
            }
            protected override void OnContinue()
            {
                foreach (var service in this.services)
                    service.Continue();
            }
        }
    }
}
